;;-----------------------------LICENSE NOTICE------------------------------------
;;  This file is part of ziloggames_z80 : An Amstrad CPC Game on Z80
;;  Copyright (C) 2022 Ander (@ander-db) / Carlos (@carlosky66) / Adrian (@RosiYo)
;;
;;  This program is free software: you can redistribute it and/or modify
;;  it under the terms of the GNU Lesser General Public License as published by
;;  the Free Software Foundation, either version 3 of the License, or
;;  (at your option) any later version.
;;
;;  This program is distributed in the hope that it will be useful,
;;  but WITHOUT ANY WARRANTY; without even the implied warranty of
;;  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;  GNU Lesser General Public License for more details.
;;
;;  You should have received a copy of the GNU Lesser General Public License
;;  along with this program.  If not, see <http://www.gnu.org/licenses/>.
;;------------------------------------------------------------------------------
.include "ai.h.s"
.include "man/entity.h.s"
.include "man/animations.h.s"
.include "physics.h.s"
.include "utils.h.s"

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; BEHAVIOURS                                                             ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Updates entity to move closer to target location
;; INPUT:
;;      A - Source coordinate value
;;      D - Target coordinate value
;; DESTROYS: A, BC, D
sys_ai_behaviour_move_coordinate:
    SUB D
    JR  Z, _exit_movement
        JR NC, _entity_bigger_than_target
        _entity_less_than_target:
            LD A, #1
            JR _exit_movement
        _entity_bigger_than_target:
            LD A, #-1
    _exit_movement:
RET

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Updates entity to move closer to target location
;; INPUT:
;;     IX - Entity memory adress
;;      B - Y target coordinate
;;      C - X target coordinate
;; DESTROYS: A, BC, D
sys_ai_behaviour_move_to:
    LD   A, X_ENTITY_POS(IX)
    LD   D, C
    CALL sys_ai_behaviour_move_coordinate
    LD   C, A

    LD   A, Y_ENTITY_POS(IX)
    LD   D, B
    CALL sys_ai_behaviour_move_coordinate
    LD   B, A

    CALL sys_physics_apply_force
RET

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Updates entity to move closer to player
;; INPUT:
;;     IX - Entity memory adress
;; DESTROYS: A, BC, D
sys_ai_behaviour_move_to_player:
    ld a, #anim_status_viser_detected
    ld ANIM_STATUS(ix), a
    CALL sys_ai_update_player_position
    CALL sys_ai_behaviour_move_to
RET

sys_ai_behaviour_wait_player:
    CALL sys_ai_update_player_position
    LD   A, Y_ENTITY_POS(IX)
    CP   B
    JR  NC, entity_not_at_player_level
        LD HL, #sys_ai_behaviour_move_to_player
        LD BEHAVIOUR_POS_H(IX), H
        LD BEHAVIOUR_POS_L(IX), L
    entity_not_at_player_level:
RET

sys_ai_behaviour_change_state:
    LD A, VY_MAX_POS(IX)
    CP #1
    JR NZ, _not_toggle
        LD  A, VX_MAX_POS(IX)
        XOR #1
        LD  VX_MAX_POS(IX), A
        LD  A, TYPE_ENTITY_POS(IX)
        XOR #E_TYPE_COLLISIONABLE
        LD  TYPE_ENTITY_POS(IX), A
        LD  A, #LASER_TOGGLE_CYCLES
        LD  VY_MAX_POS(IX), A
        JR  _exit_change_state
    _not_toggle:
        DEC A
        LD  VY_MAX_POS(IX), A
    _exit_change_state:
RET

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Jumps towards target if target not directly on top
;; INPUT:
;;     IX - Entity memory adress
;;      B - Entity type
;;      C - Target X
;; DESTROYS: AF, BC
sys_ai_behaviour_jump_to:
    LD  A, X_ENTITY_POS(IX)
    CP  C
    JR  Z, end_jump
        LD   C, #1
        JR   C, left_jump
            LD  C, #-1
        left_jump:

        LD  A, #1                            ; Retrieve As value
        ld (player_in_air), a                  ; Set flag to in the air, it cannot jump again

        LD   B, #JUMP_FORCE
        CALL sys_physics_apply_force
    end_jump:
RET

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; If enemy is able to jump, performs jump action
;; INPUT:
;;     IX - Entity memory adress
;; DESTROYS: AF, BC
sys_ai_behaviour_jump_to_player:
    CALL sys_ai_get_player_position

    ld a, (player_in_air)
    cp #0
    JR  nz, cannot_jump
        CALL sys_ai_behaviour_jump_to
    cannot_jump:
RET

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;      Ground Enemies


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Updates patrol routine, changing route when end is met
;; INPUT:
;;     IX - Entity memory adress
;;      D - Route end X
;; DESTROYS: AF, D
;; RETURNS:
;;      A - Entity X position
;;      D - New route end X
sys_ai_behaviour_update_patrol_route::
    LD   A, VX_ENTITY_POS(IX)
    CP   #0
    RET  Z
    
    AND  #0x80
    LD   A, X_ENTITY_POS(IX)
    JR   Z, _positive_patrol
    _negative_patrol:
        CP  D
        JR  Z,  _skip_comp
        JR  NC, _continue_patrol
            _skip_comp:
            ADD  A, VY_MAX_POS(IX)
        JR _exit_patrol_change
    _positive_patrol:
        CP  D
        JR  C, _continue_patrol
            SUB  A, VY_MAX_POS(IX)
    _exit_patrol_change:
    LD   VX_MAX_POS(IX), A
    LD   D, A

    RET
    _continue_patrol:
RET

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Turtle patrol behaviour, goes from A to B repeatly
;; INPUT:
;;     IX - Entity memory adress
;; DESTROYS: AF, D
sys_ai_behaviour_turtle:
    LD   D, VX_MAX_POS(IX)               ; VXMAX for turtle stands for target X
    CALL sys_ai_behaviour_update_patrol_route
    LD   A, X_ENTITY_POS(IX)
    CALL sys_ai_behaviour_move_coordinate
    LD   VX_ENTITY_POS(IX), A
RET

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Sets entity array memory position
;; INPUT:
;;      IX = Entity array, where first position stands for player
;; DESTROYS: BC
sys_ai_init:
    CALL man_entity_getArrayEnemies
    LD   (_ai_entity_call), IX
    CALL sys_ai_update_player_position
RET

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Apply behaviour to passed entity
;; INPUT:
;;      IX = Memory Adress to AI entity apply behaviour
;; DESTROYS: DE, IX
sys_ai_update_one_entity:
    LD   L, BEHAVIOUR_POS_L(IX)
    LD   H, BEHAVIOUR_POS_H(IX)
    LD   BC, #_exit_behaviour_call
    PUSH BC
    JP   (HL)
    _exit_behaviour_call:
RET



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Updates all AI entities
;; INPUT:
;; DESTROYS: HL, IX
sys_ai_update:
    CALL man_entity_getNumEnemies
    CP   #0
    RET  Z

    ai_loop:
        _artificial_counter = .+1
        LD   D, #0

        _ai_entity_call = .+2
        LD   IX, #0x0000

        LD   A, TYPE_ENTITY_POS(IX)
        CP   #E_TYPE_INVALID
        JR   NZ, valid_entity
            CALL man_entity_getArrayEnemies
            LD   (_ai_entity_call), IX
            XOR  A
            LD   (_artificial_cycle), A

            INC  D
            LD   A, D
            CP   #CYCLE_UPDATE_COUNT
            JR   C, skip_counter_reset
                XOR A
            skip_counter_reset:
            LD   (_artificial_counter), A
            RET
        valid_entity:

        _artificial_cycle = .+1
        LD   A, #0

        CP D
        JR NZ, _not_current_cycle
            CALL sys_ai_update_one_entity
        _not_current_cycle:

        INC  A
        LD   (_artificial_cycle), A
        CP   #CYCLE_UPDATE_COUNT
        JR   C, skip_cycle_reset
            XOR A
        skip_cycle_reset:
        LD   BC, #ENTITY_SIZE
        ADD  IX, BC

        LD   (_ai_entity_call), IX
    JR   ai_loop
RET

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Updates the position of the player, AI will 
;; focus over player position
;; INPUT:
;; DESTROYS: IX, BC
sys_ai_update_player_position:
    ; _entity_array = .+2
    PUSH IX
    CALL man_entity_getArray
    LD   B, Y_ENTITY_POS(IX)
    LD   C, X_ENTITY_POS(IX)
    POP IX
    ; LD (_player_location), BC
RET

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Gets player location to target
;; INPUT:
;; DESTROYS: BC
sys_ai_get_player_position:
    _player_location = .+1
    LD BC, #0x0000
RET
