;;-----------------------------LICENSE NOTICE------------------------------------
;;  This file is part of ziloggames_z80 : An Amstrad CPC Game on Z80
;;  Copyright (C) 2022 Ander (@ander-db) / Carlos (@carlosky66) / Adrian (@RosiYo)
;;
;;  This program is free software: you can redistribute it and/or modify
;;  it under the terms of the GNU Lesser General Public License as published by
;;  the Free Software Foundation, either version 3 of the License, or
;;  (at your option) any later version.
;;
;;  This program is distributed in the hope that it will be useful,
;;  but WITHOUT ANY WARRANTY; without even the implied warranty of
;;  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;  GNU Lesser General Public License for more details.
;;
;;  You should have received a copy of the GNU Lesser General Public License
;;  along with this program.  If not, see <http://www.gnu.org/licenses/>.
;;------------------------------------------------------------------------------
.include "cpctelera.h.s"
.include "physics.h.s"
.include "man/entity.h.s"
.include "utils.h.s"
.include "input.h.s"
.include "collisions.h.s"
.include "man/tilemap.h.s"
.include "render.h.s"

sys_physics_init:

RET

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Apply force in X component
;; INPUT:
;;      IX = Memory Adress entity
;;       C = X acc component
;; DESTROYS:
sys_physics_apply_force_X:
    ;; Acumulate force in X axis
    LD   A, AX_ENTITY_POS(IX)
    ADD  C
    LD   AX_ENTITY_POS(IX), A
RET

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Apply force in Y component
;; INPUT:
;;      IX = Memory Adress entity
;;       B = Y acc component
;; DESTROYS:
sys_physics_apply_force_Y:
    ;; Acumulate force in Y axis
    LD   A, AY_ENTITY_POS(IX)
    ADD  B
    LD   AY_ENTITY_POS(IX), A
RET

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Apply force, accumulting acceleration for an entity
;; INPUT:
;;      IX = Memory Adress entity
;;       C = X acc component
;;       B = Y acc component
;; DESTROYS:
sys_physics_apply_force:
    CALL sys_physics_apply_force_X
    CALL sys_physics_apply_force_Y
RET

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Sets acceleration components of the entity to zero
;; INPUT:
;;      IX = Memory Adress entity
;; DESTROYS:
sys_physics_reset_acceleration:
    LD   AX_ENTITY_POS(IX), #0
    LD   AY_ENTITY_POS(IX), #0
RET

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Applies gravity factor to an entity
;; INPUT:
;;      IX = Memory Adress entity
;; DESTROYS: AF, BC
sys_physics_apply_gravity:
    LD   A, TYPE_ENTITY_POS(IX)
    AND  #E_TYPE_GRAVITATE
    JR   Z, ingravitated_entity
        LD   B, #GRAVITY_PULL
        LD   C, #0
        CALL sys_physics_apply_force
    ingravitated_entity:
RET

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Apply current acceleration to current velocity
;; INPUT:
;;      IX = Memory Adress entity
;; DESTROYS: AF, BC
sys_physics_apply_acceleration:
    ;; Calculate new velocity in X axis
    LD   A, AX_ENTITY_POS(IX)
    LD   B, VX_ENTITY_POS(IX)
    ADD  B
    ;; Set new velocity in X axis
    LD   E, VX_MAX_POS(IX)
    CALL limit_value
    LD   VX_ENTITY_POS(IX), A

    ;; Calculate new  velocity in Y axis
    LD  A, AY_ENTITY_POS(IX)
    LD  B, VY_ENTITY_POS(IX)
    ADD B
    ;; Set new velocity in Y axis
    LD   E, VY_MAX_POS(IX)
    CALL limit_value
    LD  VY_ENTITY_POS(IX), A
RET

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Apply current velocity to current location
;; INPUT:
;;      IX = Memory Adress entity
;; DESTROYS: AF, BC
sys_physics_apply_velocity::
    ;; Calculate new X
    LD   A, VX_ENTITY_POS(IX)
    LD   B, X_ENTITY_POS(IX)
    LD   PREV_PTR_POS_L(IX), B      ; Set entity las position in X
    ADD  B

    ;; Set new X
    cp #LIMITE_LATERAL_IZQUIERDO
    jp p, #_apply_velocity_left_not_crossed
        ld a, #LIMITE_LATERAL_IZQUIERDO
        jr #_apply_velocity_right_not_crossed
    _apply_velocity_left_not_crossed:

    cp #LIMITE_LATERAL_DERECHO    
    jp m, #_apply_velocity_right_not_crossed
        ld a, #LIMITE_LATERAL_DERECHO
    _apply_velocity_right_not_crossed:

    LD   X_ENTITY_POS(IX), A

    ;; Calculate new Y
    ld   a, Y_ENTITY_POS(ix)
    ld   PREV_PTR_POS_H(ix), a
    LD   A, VY_ENTITY_POS(IX)
    LD   B, Y_ENTITY_POS(IX)
    ADD  B
    LD B, A                     ; Store new pos

    CP #LIMITE_SUPERIOR
    JP P, #_not_lim_sup
        LD A, #LIMITE_SUPERIOR
    _not_lim_sup:

    LD  A, TYPE_ENTITY_POS(IX)
    AND #E_TYPE_PLAYER
    LD  A, B                    ; Retrieve new pos
    JR  Z, move_entity_y
        cp #CAMARA_SCROLL_THRESHOLD    
        jp m, move_entity_y

        bit 7, a
        jr nz, move_entity_y

        push ix
        ld a, #0x01
        call sys_render_set_apply_scroll 
        ld a, #0x01
        call sys_tilemap_update
        pop ix
        LD  A, Y_ENTITY_POS(IX)
        sub #0x08
        ld PREV_PTR_POS_H(ix), a
        RET
    move_entity_y:

    LD   Y_ENTITY_POS(IX), A
RET

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Physics calculations for one entity
;; INPUT:
;;      IX = Memory Adress entity
;; DESTROYS: AF, BC
sys_physics_update_one_entity:
    CALL sys_physics_apply_gravity
    CALL sys_physics_apply_acceleration
    ; CALL sys_physics_check_collision_flags
    CALL sys_physics_apply_velocity
    CALL sys_physics_reset_acceleration
RET

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Physics calculations for all entities
;; INPUT:
;; DESTROYS: HL, BC,
sys_physics_update:
    LD   HL, #sys_physics_update_one_entity
    LD    A, #E_TYPE_MOVABLE
    CALL man_entity_forall_matching
RET
